"""Definition of all commands available in ReColonyzer."""
import logging

from recolonyzer import analysis, filesystem, utils
from recolonyzer.commands import abstract

logger = logging.getLogger(__name__)


class AnalyseCommand(abstract.AbstractCommand):

    _SUBCOMMAND = 'analyse'
    _DESCRIPTION = """\
    Analyse timeseries of QFA images: locate cultures on
    plate, segment image into agar and cells, apply lighting correction,
    and generate output files for each image.
    """

    def register_arguments(self, parser):
        parser.add_argument(
            "-d",
            "--directory",
            type=str,
            help="""Directory in which to search for image files that need to be
            analysed.
            Default: current directory.""",
            default=".")
        parser.add_argument(
            "-c",
            "--light_correction_off",
            help="""Disables lighting correction between images.
            Default: Light correction is enabled.""",
            action="store_false")
        parser.add_argument(
            "-q",
            "--quiet",
            help="""Suppresses messages printed during the analysis.
            Default: show messages.""",
            action="store_true")
        parser.add_argument(
            "-e",
            "--endpoint",
            help="""Analyses only the final image in the series.
            It is useful to test single images.
            Default: False. Analyse all images in the directory.""",
            action="store_true")
        parser.add_argument(
            "-g",
            "--grid_format",
            type=str,
            nargs='+',
            help="""Specifies rectangular grid format.
            Important: specify number of rows and number of columns, in this order
            (e.g. -g 8x12 or -g 8 12).
            Default: 8x12.""",
            default=['8x12'])
        parser.add_argument(
            "-f",
            "--fraction",
            type=utils.range_float,
            help="""Specifies the minimum fraction of the image that corresponds
            to the grid. Adjust it if grid occupies a small part of the total image.
            Default: 0.8.""",
            default=0.8)

    def run(self, args):
        # Setup logger
        if args.quiet:
            logging.basicConfig(format="%(message)s", level=logging.INFO)
        else:
            logging.basicConfig(format="%(message)s", level=logging.DEBUG)

        # Print information of inputs to users.
        utils.summarise(args)

        # Get working directory.
        fdir = filesystem.get_directory(args.directory)
        nrow, ncol = utils.get_grid_format(args.grid_format)

        # Create needed directories to save outputs.
        filesystem.arrange_directories(fdir)
        # Obtain list of images to analyse.
        imanalyse = filesystem.get_images(fdir, args.endpoint)

        # Perform main logic.
        output_dfs, output_images = analysis.analyse_timeseries_qfa(
            imanalyse,
            nrow,
            ncol,
            light_correction=args.light_correction_off,
            fraction=args.fraction)

        # Save outputs.
        file_names = [filesystem.get_file_name(f) for f in imanalyse]
        filesystem.save_outputs(file_names, output_dfs, output_images, fdir)

        logger.info("No more images to analyse. I'm done")



import abc


class AbstractCommand:
    """Abstract class used to define shared logic for all commands in
    recolonyzer."""

    # Name for the command.
    _SUBCOMMAND = None
    # Description of the subcommand shown in the help.
    _DESCRIPTION = None

    def register_parser(self, subparsers):
        # Register this class as a subcommand of the main recolonyzer script.
        group = subparsers.add_parser(
            self._SUBCOMMAND,
            help=self._DESCRIPTION,
            description=self._DESCRIPTION)
        # Call the method of the class responsible for adding all arguments
        # given by the user.
        self.register_arguments(group)
        # Register the method `run()` to be called when a subcommand is called.
        group.set_defaults(func=self.run)

    @abc.abstractmethod
    def register_arguments(self, parser):
        """Register command arguments into the given parser."""

    @abc.abstractmethod
    def run(self, args):
        """Run command logic."""



import datetime
import glob
import os

from recolonyzer.commands import abstract


class RenameImagesCommand(abstract.AbstractCommand):

    _SUBCOMMAND = 'rename_images'
    _DESCRIPTION = """\
    Rename image files according to date time format. By default it only shows
    the renaming of the image files, without applying the modification. Use the
    flag --no_dry_run to apply the renaming.
    """

    def register_arguments(self, parser):
        parser.add_argument(
            "-d",
            "--directory",
            help="""Directory where the images are stored.
            Default: current directory.""",
            default=".")
        parser.add_argument(
            "-g",
            "--glob",
            help="""Glob to match all images in the given directory.
            Default: *.jpg (matches all jpg files).""",
            default="*.[Jj][Pp][Gg]")
        parser.add_argument(
            "-p",
            "--prefix",
            help="File prefix to generate. Default: QFA_90000000001_.",
            default="QFA_90000000001_")
        parser.add_argument(
            "-s",
            "--start_time",
            help="""Initial time for files renaming.
            Required format: 'YYYY-MM-DD hh:mm'. Default: now.""",
            type=lambda s: datetime.datetime.strptime(s, '%Y-%m-%d %H:%M'),
            default=datetime.datetime.now())
        parser.add_argument(
            "-i",
            "--interval",
            help="Time interval between files in minutes. Default: 30.",
            type=int,
            default=30)
        parser.add_argument(
            "--no_dry_run",
            help="""By default the change of names won't take place.
            If you specify this flag the renaming will take place.""",
            action="store_true")

    def run(self, args):
        # Get list of files to rename.
        file_names = glob.glob(os.path.join(args.directory, args.glob))
        file_names.sort()

        if not file_names:
            raise RuntimeError(
                'No file names found with the glob expression: {}'.format(
                    args.glob))
        print('Found {} files to rename.'.format(len(file_names)))

        # Get extension of the files.
        extension = os.path.splitext(file_names[0])[1]

        # Define starting date time and time interval.
        current_datetime = args.start_time
        delta_time = datetime.timedelta(minutes=args.interval)

        for file_path in file_names:
            # Create new file name.
            new_file_name = args.prefix + current_datetime.strftime(
                "%Y-%m-%d_%H-%M-%S") + extension
            new_path = os.path.join(os.path.dirname(file_path), new_file_name)

            # Show the renaming to the user.
            print("{} -> {}".format(file_path, new_path))
            # If specified by the user, perform the actual renaming of the file.
            if args.no_dry_run:
                os.rename(file_path, new_path)

            current_datetime += delta_time

        if args.no_dry_run:
            print("All files renamed!")



"""Functions to compute the outputs that will be saved.
"""
import logging
import os
import time

import cv2
import numpy as np
import pandas as pd

from recolonyzer import filesystem, image_processing

logger = logging.getLogger(__name__)


def analyse_timeseries_qfa(images_paths,
                           nrow,
                           ncol,
                           light_correction=False,
                           fraction=0.8):
    # Set timer
    start_time = time.time()

    # Obtain first and last image
    latest_image = images_paths[-1]
    earliest_image = images_paths[0]

    # Print information to users
    logger.debug("")
    logger.debug("Starting analysis:")
    logger.debug("Earliest image: %s", earliest_image)
    logger.debug("Latest image: %s", latest_image)
    logger.debug("")
    logger.debug("Computing position of the grid")
    logger.debug("This may take a few seconds...")
    logger.debug("")

    # Get latest image to detect culture locations
    im_n = cv2.imread(latest_image, cv2.IMREAD_GRAYSCALE)

    # Perform some normalization in order to remove outliers and noise.
    # The 1% and 99% quantile clipping is a good option.
    quants = np.quantile(im_n.ravel(), [0.01, 0.99])
    im_n = np.clip(im_n, *quants)
    im_n = np.array(im_n, dtype=np.uint8)

    _, min_loc, pat_h, pat_w = image_processing.get_position_grid(
        im_n, nrow, ncol, fraction)

    # Cut the original image with the size of the best pattern match.
    w_right = int(min_loc[0] + pat_w)
    h_bottom = int(min_loc[1] + pat_h)
    im_ = im_n[min_loc[1]:h_bottom, min_loc[0]:w_right]

    # Find spots and agar based on an automatic threshold
    _, mask = cv2.threshold(
        np.array(im_, dtype=np.uint8), 0, 255,
        cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU)

    # Locate the position of the spots and the agar into different masks.
    grd = np.ones(mask.shape, dtype=bool)
    spots = np.logical_and(grd, ~mask)  # grd & ~mask
    agar = np.logical_and(grd, mask)  # grd & mask

    # Reset mask to avoid problems in future iterations.
    mask = None

    output_dfs = []
    output_images = []
    logger.debug("Analysing each of the images:")
    for file_name in images_paths:

        img = cv2.imread(file_name, cv2.IMREAD_GRAYSCALE)
        arr = img[min_loc[1]:h_bottom, min_loc[0]:w_right]

        # Set threshold depending on last image
        thresh, _ = cv2.threshold(
            np.array(arr, dtype=np.uint8), 0, 255,
            cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU)
        # If the threshold is lower than the values of the agar, set it higher.
        thresh = max(thresh, np.mean(arr[agar]) + 1)

        # Create mask to detect spots in each image.
        # This will be used only to compute the area of the spots.
        mask = np.ones(arr.shape, dtype=np.bool)
        mask[arr < thresh] = False

        # Measure culture phenotypes.
        output_dfs.append(
            measure_outputs(arr, mask, pat_h, pat_w, nrow, ncol, file_name,
                            spots, light_correction))

        # Save mask for a visual check.
        output_images.append(mask.astype(np.uint8) * 255)

        # Reset mask to avoid problems in the next iteration.
        mask = None
        logger.debug("Analysis complete for %s", os.path.basename(file_name))
    logger.debug("All analyses finished in {:.2f} seconds".format(time.time() -
                                                                  start_time))
    return output_dfs, output_images


def measure_outputs(im, mask, pat_h, pat_w, nrow, ncol, file_name, spots,
                    correction):
    """Add intensity measures and other measurements to a final dictionary.
    This dictionary will be outputed as a data-frame.
    """
    # Outputs will be saved in a dictionary with the following features.
    allrows, allcols, allintensities, allareas = [], [], [], []
    allcolonymeans, allcolonyvariance, allbackgroundmeans = [], [], []

    # Extract the windows for each spot. First get the window widht and heigh.
    d_x = int(pat_h / nrow)
    d_y = int(pat_w / ncol)

    # Maximum intensity that we can see in a grayscale image is 255.
    int_max = 255

    # From now on: reference point is the top left corner of the rectangle.
    for i in range(nrow):
        for j in range(ncol):
            # Compute position x of each window or patch
            p_x = j * d_x  # Position x = ncol x dimension_x of window (d_x)
            p_y = i * d_y  # Position y = nrow x dimension_y of window (d_y)
            # Get window of this specific colony
            patch = im[p_y:p_y + d_y, p_x:p_x + d_x]
            mask_patch = mask[p_y:p_y + d_y, p_x:p_x + d_x]
            spots_patch = spots[p_y:p_y + d_y, p_x:p_x + d_x]

            # Perform some normalization in order to remove outliers and noise.
            # The 1% and 99% quantile clipping is a good option.
            quants = np.quantile(patch.ravel(), [0.01, 0.99])
            patch = np.clip(patch, *quants)

            # Compute area of colony based on the mask_patch
            area_patch = np.sum(mask_patch) / patch.size
            # Compute background mean of patch that will be useful to correct
            # for differences in intensity between AND within images
            if correction:
                backgr = np.mean(patch[~spots_patch])
            else:
                backgr = 0

            # Compute tiles intensities and filter only according to spots
            tile = patch - backgr
            spot_intensities = tile[spots_patch]

            # If there are colonies in the window or patch, compute metrics
            if spot_intensities.size > 1:
                # Compute the mean of the intensity (Normalized from 0 to 1)
                colony_mean_patch = np.mean(spot_intensities / int_max)
                # Compute the variance of the intensity
                colony_variance_patch = np.var(spot_intensities / int_max)
            else:  # If there aren't colonies in the window, intensities are 0
                colony_mean_patch = colony_variance_patch = 0

            # Compute all intensity values normalized by the size of window
            intens_patch = np.sum(tile) / (tile.size * int_max)

            # Compute agar information (agar is the opposite to the mask)
            bkgrnd = tile[~spots_patch]

            # If there is background in the patch, compute mean of intensities
            if bkgrnd.size > 1:
                background_mean_patch = np.mean(bkgrnd / int_max)
            else:  # If there is no background in the patch, intensities are 0
                background_mean_patch = 0

            # Save all variables in the final lists
            allrows.append(i + 1)
            allcols.append(j + 1)
            allintensities.append(intens_patch)
            allareas.append(area_patch)
            allcolonymeans.append(colony_mean_patch)
            allcolonyvariance.append(colony_variance_patch)
            allbackgroundmeans.append(background_mean_patch)

    # Save final outputs
    fname = filesystem.get_file_name(file_name)
    return pd.DataFrame({
        "Row": allrows,
        "Column": allcols,
        "Intensity": allintensities,
        "Area": allareas,
        "ColonyMean": allcolonymeans,
        "ColonyVariance": allcolonyvariance,
        "BackgroundMean": allbackgroundmeans,
        "Barcode": [fname[0:15]] * len(allrows),
        "Filename": [fname] * len(allrows),
    })



"""Functions to get the agar and spot colors and to locate the grid.
"""

import itertools

import cv2
import numpy as np
from scipy.signal import find_peaks
from tqdm import tqdm


def get_agar_spot_color(im):
    """Get the color value of the agar and the spot.
    Obtain the values by checking the largest two peaks in the color histogram.
    """
    im_ = np.asarray(im).ravel()

    # Compute histogram of intensities
    bincnt = np.bincount(im_)

    # Find peaks in histogram
    peaks, _ = find_peaks(bincnt, width=1)
    vals = bincnt[peaks]
    idx = np.argsort(vals)[::-1]

    # Take first peak: this will be the agar
    color_agar = peaks[idx[0]]
    # Take second peak: these will be the spots
    color_spot = peaks[idx[1]]
    return int(color_agar), int(color_spot)


def get_position_grid(im, nrow, ncol, frac):
    """Get the position of the grid in the image.
    Use an artificial pattern and the normalized squared difference to find it.
    Return the position of the grid and the size of the scaled the pattern.
    """
    # Find color values of agar and spots
    color_agar, color_spot = get_agar_spot_color(im)
    # Size of the given image.
    h, w = im.shape

    #Create the pattern of circles following the grid structure.
    rpix = int(h / nrow)  # Set random size of window that will later be scaled
    rspot = int((rpix * 0.7) / 2)  # Set random size of spot radius to scale
    pattern = np.ones((nrow * rpix, ncol * rpix), dtype=np.uint8) * color_agar
    for i, j in itertools.product(range(nrow), range(ncol)):
        cv2.circle(pattern,
                   (j * rpix + int(rpix / 2), i * rpix + int(rpix / 2)), rspot,
                   color_spot, -1)

    # Check for structure between frac and 100% of the total image size.
    min_fraction, max_fraction = frac, 1.0
    # Check for structure using iterations by increasing 0.2% of the image
    iterations = int((max_fraction - min_fraction) * 100 / 0.2)

    # Store the info about the best scale where pattern fits the image.
    found = None
    for fraction in tqdm(
            np.linspace(min_fraction, max_fraction, iterations),
            desc='Fitting pattern...'):
        # Scale the pattern to the fraction of the given image.
        pat_h, pat_w = int(h * fraction), int(w * fraction)
        pattern_scaled = cv2.resize(pattern, (pat_w, pat_h))
        # Match the scaled pattern with the image.
        res = cv2.matchTemplate(im, pattern_scaled, cv2.TM_SQDIFF_NORMED)
        # The function gives the map of squared difference, so optimal location
        # will be the one with minimum value.
        min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(res)
        if found is None or min_val < found[0]:
            found = (min_val, min_loc, pat_h, pat_w)
    return found



"""Functions to prepare de directories and to obtain the files for analysis.
"""

import logging
import os

import cv2

logger = logging.getLogger(__name__)


def get_directory(directory=None):
    """Get directory to work with."""
    # Set variable fdir = current directory, if user didn't specify another dir
    if not directory:
        fdir = os.getcwd()
    # Set variable fdir = directory chosen by the user, if a dir is specified
    else:
        fdir = os.path.realpath(os.path.expanduser(directory))
        # Make sure that the directory exists. Otherwise, print error and exit
        if not os.path.isdir(fdir):
            raise ValueError("Directory doesn't exist. Check --directory.")
    return fdir


def arrange_directories(directory):
    """Creates directories to save output files."""
    # Set directories to keep or create.
    imdir = os.path.join(directory, "Output_Images")
    datdir = os.path.join(directory, "Output_Data")

    # Create directories to save the outputs.
    if not os.path.isdir(imdir):
        os.mkdir(imdir)
    if not os.path.isdir(datdir):
        os.mkdir(datdir)
    logger.debug("Outputs will be saved in " + imdir + " & " + datdir)


def get_images(directory, endpoint):
    """Get filenames for all images in current directory.
    Return a list of filenames to analyse"""

    # Set image formats that will be considered.
    image_formats = {".jpg", ".jpeg", ".tif", ".tiff", ".png"}
    # Obtain all file names in the specified directory
    filenames = os.listdir(directory)

    # Take only the images from the directory (images have any image_formats)
    allfiles = [
        newfile for newfile in filenames
        if os.path.splitext(newfile.lower())[1] in image_formats
    ]
    allfiles.sort()

    # If endpoint is True, then take only first and last images
    if endpoint:
        allfiles = [allfiles[-1]]

    # Obtain all images to analyse
    imanalyse = []
    for filename in allfiles:
        fullfilename = os.path.join(directory, filename)
        imanalyse.append(fullfilename)

    # If there aren't any new images to analyse, display error and exit
    if not imanalyse:
        raise ValueError("No new images to analyse in " + directory + ".")
    imanalyse.sort()
    return imanalyse


def get_file_name(image_path):
    """Get file names from given path removing extension and directories."""
    return os.path.basename(os.path.splitext(image_path)[0])


def save_outputs(file_names, output_dfs, output_images, output_base_dir):
    """Saves outputs."""
    logger.debug("Saving outputs...")

    for file_name, output_df, output_image in zip(file_names, output_dfs,
                                                  output_images):

        # Save DataFrame of output metrics.
        output_df.to_csv(
            os.path.join(output_base_dir, "Output_Data",
                         "{}.out".format(file_name)),
            sep="\t",
            index=False)
        # Save Image mask
        img_outputs_dir = os.path.join(output_base_dir, "Output_Images",
                                       "{}.png".format(file_name))
        cv2.imwrite(img_outputs_dir, output_image)
